/* $Id: build.c,v 1.9 2000/01/15 18:00:44 alinden Exp $ */

/*
*  File             : build.c
*  Written by       : alinden@gmx.de
*  Portability      : Should be POSIX compliant  
*  Copyright:       : GPL
*
*  DESCRIPTION: Contains the stuff needed to build the crontab database.
*  This is done by calling read_config(). The list is linked three times for
*  user, entry and Q commands. To be memory and CPU efficient, the time fields are
*  stored as a bit oriented index within this struct. For example, the minute field
*  is an array of 60 bits in 8 bytes where each bit corresponds to a minute (0-59). 
*/

#include "conf.h"
#include "crond.h"
#include "extern.h"

typedef struct   /* crontab options */
{
	char *user;
	char *mailto;
	char *subject;
	int  htmout,atmout;
	char dopt,bopt,lopt,ropt,mopt,qopt;
} CTOPTS;

void msgout(char *); 
void log_info(char *);
void log_warn(char *);
char *cpstr(size_t, char *, ...);
void rmenv(NVR *);
void rmcstr(CONFIG *);
void rmtbuf(CONFIG *);
int create_socket(char *sockname);
time_t runtab(CONFIG *, time_t);
int chk_tmstr(char *, unsigned char *, int);
void lbnm(char *, int);
void flushqueue(void);
void flushqctab(char *);

unsigned long uhours(char **inbuf)   /* returns number of minutes for U option */ 
{                                    /* inbuf: field array */                
	unsigned long tminutes,timeval;

	tminutes = timeval = 0;       /* minute field */
	if ( *inbuf[0] != '*' && sscanf(inbuf[0], "%lu", &timeval) < 1 )
		return(0);
	tminutes = timeval;

	timeval = 0;                 /* hour field */           
	if ( *inbuf[1] != '*' && sscanf(inbuf[1], "%lu" , &timeval ) < 1 )
		return(0);
	tminutes += timeval*60;

	timeval = 0;                 /* day field */
	if ( *inbuf[2] != '*' && sscanf(inbuf[2], "%lu", &timeval) < 1 )
		return(0);
	tminutes += timeval*24*60;

	timeval = 0;                 /* week field */
	if ( *inbuf[3] != '*' && sscanf(inbuf[3], "%lu", &timeval) < 1 )
		return(0);
	tminutes += timeval*24*60*7;
	
	return(tminutes);
}

int rmuser(char *ctab)  /* remove a user */
{
	CONFIG *tbuf,*pbuf,*puser;
	int nentries=0,suser=0;
	
	if(strcmp(ctab, NCTROOT) == 0)
		ctab = NULL;

	tbuf = pbuf = puser = cblock;

	while(tbuf != NULL)  /* find user */
	{
		if((ctab == NULL && tbuf->cmain) ||
		  (ctab != NULL && ! tbuf->cmain && strcmp(ctab, tbuf->username) == 0))
			break;

		puser = tbuf;
		suser = 1;
		tbuf = tbuf->nuser;
	}
	
	if(tbuf == NULL)
		return(0);
		
	flushqctab(ctab);
	rmcstr(tbuf);

	do   /* remove entries for user */
	{
		pbuf = tbuf;
		tbuf = tbuf->nblock;
		rmtbuf(pbuf);
		nentries++;
		
	} while(tbuf != NULL && tbuf->stblock == 0);

	/* close the chain */

	if(! suser)
		cblock = tbuf;
	else
	{
		puser->nuser = tbuf;

		do
		{
			pbuf = puser;
			puser = puser->nblock;
		
		} while(puser != NULL && puser->stblock == 0);

		pbuf->nblock = tbuf;
	}	
	
	return(nentries);
}

void rmenv(NVR *nvrbuf)  /* remove environment for user */
{
	NVR *pvrbuf;
	
	while(nvrbuf != NULL)
	{
		pvrbuf = nvrbuf;
		nvrbuf = nvrbuf->nblock;
		free(pvrbuf);
	}
}			

void rmcstr(CONFIG *tbuf)  /* remove strings and environment */
{
	NVR *nnvr=NULL;
	
	do
	{
		if(tbuf->envr != NULL && tbuf->envr != nnvr)
		{
			nnvr = tbuf->envr;
			free(nnvr);
			tbuf->envr = NULL;
		}
		tbuf = tbuf->nblock;
	}
	while(tbuf != NULL && ! tbuf->stblock);
}

void rmtbuf(CONFIG *tbuf) /* expire block */
{
	if(tbuf->pid)
	{
		tbuf->nblock = jblock;
		jblock = tbuf;
	}
	else if(tbuf->qdelay || tbuf->swpdelay)
	{
		tbuf->nblock = qblock;
		qblock = tbuf;
	}
	else free(tbuf);
}

void rmqjobs(void)
{
	CONFIG *nbuf,*pbuf;
	
	nbuf = qblock;
	while(nbuf != NULL)  /* remove old jobs */
	{
		pbuf = nbuf;
		nbuf = nbuf->nblock;
		free(pbuf);
	}
	qblock = NULL;
}	

int adduser(CONFIG *tbuf)  /* add a user */
{
	CONFIG *pbuf=NULL,*nbuf,*ostart;
	int nentries=0;
	
	if(tbuf->cmain || strcmp(tbuf->username, "root") == 0)  /* priority */
	{
		ostart = cblock;
		cblock = tbuf;
		tbuf->nuser = ostart;
	}	
	else
	{
		ostart = NULL;
		
		if(cblock == NULL)
			cblock = tbuf;
		else
		{
			for(pbuf=cblock; pbuf != NULL; pbuf=pbuf->nuser)
				nbuf=pbuf;
				
			nbuf->nuser = tbuf;
			
			while(nbuf != NULL)
			{
				pbuf = nbuf;
				nbuf = nbuf->nblock;
			}
			pbuf->nblock = tbuf;
		}
	}
	
	while(tbuf != NULL)
	{
		pbuf = tbuf;
		tbuf = tbuf->nblock;
		nentries++;
	}
	pbuf->nblock = ostart;
	
	return(nentries);
}

/* take instr as crontab entry and separate tokens until command is reached
   return array of fields or NULL */

char **sptoken(char *instr) 
{
	static char *ninstr[20];
	int i,i2; 		    			 
			
	if(! *instr)
		return(NULL);
		
	if((ninstr[0] = strtok(instr, "\t ")) == NULL)
		return(NULL);
		
	for(i = 1; i < 6; i++)
	{
		if((ninstr[i] = strtok(NULL, "\t ")) == NULL)
			return(NULL);
	}

	i--;

	while(*ninstr[i] == '-' || *ninstr[i] == '+')  /* options field */
	{
		if(i == 19 || (ninstr[++i] = strtok(NULL, "\t ")) == NULL)
			return(NULL);
	}

	ninstr[i+1] = NULL;

	i2 = strcspn(ninstr[i], "\n");
	
	if(*(ninstr[i]+i2) == '\n')
		*(ninstr[i]+i2) = 0;
	else
	{
		*(ninstr[i]+i2) = ' ';
		*(ninstr[i]+strcspn(ninstr[i], "\n")) = 0;   /* chop trailing \n */
	}	

	return(ninstr);
}	

/* check if crontab is owned by user, return -1 if not, 0 if user crontab, 1 if root crontab */

char userperm(char *ctab)
{					       			
	struct passwd *pword;
	char *username,cmain;
	struct stat buf;

	if(strcmp(ctab, NCTROOT) == 0)
	{
		username = "root";
		cmain = 1;
	}
	else
	{
		username = ctab;
		cmain = 0;
	}	

	while(stat(ctab, &buf) < 0)
	{
		if(errno != ENOMEM)
		{
			log_warn(cpstr(100, "userperm(): stat for %s failed", ctab));
			return(-1);
		}	

		sleep(1);	
	}

	if(! buf.st_uid)   /* owned by root */
		return(cmain);

	while((pword = getpwnam(username)) == NULL)    /* no such user */
	{
		if(errno != ENOMEM)
		{
			log_warn(cpstr(100, "ignored crontab for %s: no such user", username));
			return(-1);
		}	
		
		sleep(1);
	}	
	
	if(buf.st_uid != pword->pw_uid )  /* not owner of file */
	{	
		log_info(cpstr(100, "crontab for %s doesn't match owner", username));
		return(-1);
	}	
	return(cmain);
}

void insert_delayjob(CONFIG *tbuf)  /* check for expired job and insert data */
{
	CONFIG *nbuf;
	
	for(nbuf = qblock; nbuf != NULL; nbuf = nbuf->nblock)
	{
		if(tbuf->cmain == nbuf->cmain && 
		strcmp(tbuf->username, nbuf->username) == 0 &&
		strcmp(tbuf->cmdstr, nbuf->cmdstr) == 0)
		{
			tbuf->exite = nbuf->exite;
			tbuf->exitc = nbuf->exitc;
			tbuf->lastexec = nbuf->lastexec;

			if(tbuf->qopt)
				tbuf->qdelay = nbuf->qdelay;

			if(tbuf->atmout)
				tbuf->swpdelay = nbuf->swpdelay;
				
			if((tbuf->qopt && nbuf->qdelay) || (tbuf->atmout && nbuf->swpdelay))
				tbuf->nextexec = nbuf->nextexec;

			break;
		}
	}
}

int readopts(char **cpbuf, CTOPTS *copts, char cmain)  /* read option fields */
{
	int i,ncmds=0;
	char inp,cinv;
	
NBLV:	while((cinv = *cpbuf[ncmds] == '-') || *cpbuf[ncmds] == '+')
	{
		for(i = 1; *(cpbuf[ncmds]+i); i++)
		{
			inp = tolower(*(cpbuf[ncmds]+i));
	
			if(*(cpbuf[ncmds]+i+1) == '=')
			{
				switch(inp)
				{
					case 'u':
						if(cmain)
							copts->user = cpbuf[ncmds]+i+2;

						ncmds++;
						goto NBLV;
			
					case 'm':
						if(! *(copts->mailto = cpbuf[ncmds]+i+2))
							copts->mailto = NULL;
						else
							copts->mopt = 1;
					
						ncmds++;
						goto NBLV;

					case 'h':
						if(*(cpbuf[ncmds]+i+2))
							copts->htmout = atoi(cpbuf[ncmds]+i+2)*60;
					
						ncmds++;
						goto NBLV;

					case 'a':
						if(*(cpbuf[ncmds]+i+2))
							copts->atmout = atoi(cpbuf[ncmds]+i+2)*60;
					
						ncmds++;
						goto NBLV;

					case 's':
						copts->subject = cpbuf[ncmds]+i+2;

						if(strlen(copts->subject) >= 40)
							*(copts->subject+40) = 0;
			
						ncmds++;
						goto NBLV;
				}
			}
			else
			{
				switch(inp)
				{
					case 'd':
						copts->dopt = cinv;
						break;

					case 'b':
						copts->bopt = cinv;
						break;

					case 'h':
						copts->htmout = cinv?3*60*60:0;
						break;

					case 'l':
						copts->lopt = cinv;
						break;

					case 'r':
						copts->ropt = cinv;
						break;
						
					case 'q':	
						copts->qopt = cinv;
						break;
						
					case 'a':
						copts->atmout = cinv?10*60:0;
						break;
				}		
			}
		}	
		ncmds++;
	}
	return(ncmds);
}	

CONFIG *readctab(char *fname, CTOPTS *defopts, char cmain, int *nentries)  /* read a crontab, return first block or NULL */
{
	FILE *crontab; 
	char *ninbuf,**cpbuf,*outbuf;
	char *inds[4],*nbuf,*nptr;
	CONFIG *tbuf,*pbuf=NULL;
	CONFIG *sblock,*lstqbuf=NULL,*fqblock=NULL;
	NVR *nvrbuf,*pvrbuf=NULL,*snvr=NULL;
	char inbuf[11*1025];
	char envmto[200],envmuse=0;
	size_t slen;
	time_t ltm;
	char *sstr,ashell=0,ahome=0,anice=0;
	CTOPTS copts;
	int qtmout=qtimeout;
	unsigned long tminutes; 
	int i,dres,wres,ncmds,cline=0,ecmd=0;

	if((crontab = fopen(fname, "r")) == NULL)
		return NULL;

	sblock = NULL;

	while(time(&ltm) < 0)
		sleep(1);

	while(fgets(inbuf, 1024, crontab) != NULL)
	{
		cline++;
		
		do
		{
			for(nbuf = inbuf+strlen(inbuf); nbuf > inbuf && (*nbuf == ' ' || *nbuf == '\t' || *nbuf == '\n' || ! *nbuf); nbuf--);

			if(*nbuf != '\\')  /* no continuation */
				break;

			if(! cmain && (*nentries)++ > UTABMAX)
				break;

			if(strlen(inbuf) > 10*1024)
				break;

		} while(fgets(nbuf, 1024, crontab) != NULL);
		
		if(strchr(inbuf, '\n') == NULL)
		{
			*(inbuf+(i = strlen(inbuf))) = '\n';
			*(inbuf+i+1) = 0;
		}

		ninbuf = inbuf + strspn(inbuf, "\t ");  /* strip leading space */

		if(*ninbuf == '#' || *ninbuf == '\n')
			continue;

		/* limit user crontabs to 200 entries */
	
		if(! cmain && (*nentries)++ > UTABMAX)
			break;

		/* test for environment entry */

		outbuf = ninbuf + (i = strcspn(ninbuf, "\t\n= "));

		if (i && *outbuf == '=')
		{
			if(strncmp(ninbuf, "LOGNAME", 7) == 0 || strncmp(ninbuf, "USER", 4) == 0)
				continue;

			*(outbuf + strcspn(outbuf, "\n")) = 0;
			
			if(! envmuse && strncmp(ninbuf, "MAILTO", 6) == 0 && *(outbuf+1))
			{
				if(strlen(outbuf) > 198)
					*(outbuf+198) = 0;
			
				strcpy(envmto, outbuf+1);
				envmuse = 1;
			}
			
			while((nvrbuf = malloc(sizeof(NVR)+strlen(ninbuf)+1)) == NULL)
				sleep(1);

			strcpy((char *)nvrbuf+sizeof(NVR), ninbuf);
			nvrbuf->nblock = NULL;
			
			(snvr == NULL) ? (snvr = nvrbuf) : (pvrbuf->nblock = nvrbuf);
			pvrbuf = nvrbuf;

			if(strncmp(ninbuf, "SHELL", 5) == 0)
				ashell = 1;

			if(strncmp(ninbuf, "HOME", 4) == 0)
				ahome = 1;

			if(strncmp(ninbuf, "NICE", 4) == 0)
				anice = 1;

			if(cmain && *(outbuf+1) && strncmp(ninbuf, "QTIME", 5) == 0)
				qtmout = atoi(outbuf+1);
			
			continue;
		}

		if((cpbuf = sptoken(ninbuf)) == NULL)
			continue;

		/* make entry of cron field */

		if(*cpbuf[0] == 'U' || *cpbuf[0] == 'u')  /* U option used */ 
		{
			if((tminutes = uhours(&cpbuf[1])) == 0)
				continue;
		}
		else
		{	
			lbnm(cpbuf[3], 1);
			lbnm(cpbuf[4], 0);

			tminutes = 0;
		}	

		/* options */

		memcpy(&copts, defopts, sizeof(CTOPTS));
		ncmds = 5+readopts(&cpbuf[5], &copts, cmain);
		inds[0] = copts.user;
		inds[1] = cpbuf[ncmds];
		inds[2] = copts.subject;

		if(tminutes)
			copts.dopt = 0;
		
		if(copts.mailto == NULL)
		{
			if(envmuse)
				inds[3] = envmto;
			else
				inds[3] = NULL;
		}
		else inds[3] = copts.mailto;
		
		/* inds[0] points to username, inds[1] to cmdline, inds[2] to subject,
		   inds[3] to mailto now */

		slen = strlen(inds[0])+strlen(inds[1])+2;

		if(inds[2] != NULL)
			slen += strlen(inds[2])+1;

		if(inds[3] != NULL)
			slen += strlen(inds[3])+1;

		while((tbuf = malloc(sizeof(CONFIG)+slen)) == NULL)
			sleep(1);

		if(! tminutes)
		{	
			if ( ! chk_tmstr(cpbuf[0], tbuf->minutes, 60)
			  || ! chk_tmstr(cpbuf[1], tbuf->hours, 24)
			  || ! (dres = chk_tmstr(cpbuf[2], tbuf->days, 32))
			  || ! chk_tmstr(cpbuf[3], tbuf->month, 13)
			  || ! (wres = chk_tmstr(cpbuf[4], tbuf->weekday, 7)))
			{
			 	free(tbuf);

				if(svsock != -1 && ! ecmd++)
				{
					sstr = cpstr(100, "%s line %d: parse error\n", fname, cline);
					write(svsock, sstr, strlen(sstr));
				}
			 	continue;
			}

			if( wres == 2 )
				tbuf->wdres = 0x01;
			else
				tbuf->wdres = 0;		

			if( dres == 2 )
				tbuf->wdres |= 0x02;	
		}
			
		/* first pointer into cblock,then expand CONFIG struct */

		(sblock == NULL) ? (sblock = tbuf) : (pbuf->nblock = tbuf); 

		tbuf->username = strcpy((char *)tbuf+sizeof(CONFIG), inds[0]);
		tbuf->cmdstr = strcpy(tbuf->username+strlen(inds[0])+1, inds[1]);

		nptr = tbuf->cmdstr+strlen(inds[1])+1;
		
		if(inds[2] != NULL)
			tbuf->jobid = strcpy(nptr, inds[2]);
		else
			tbuf->jobid = NULL;

		if(inds[3] != NULL)
		{
			if(tbuf->jobid != NULL)
				nptr += strlen(tbuf->jobid)+1;

			tbuf->mailto = strcpy(nptr, inds[3]);
		}
		else tbuf->mailto = NULL;

		if(copts.qopt)  /* link Q commands */
		{
			if(lstqbuf != NULL)
				lstqbuf->nqblock = tbuf;

			lstqbuf = tbuf;

			if(fqblock == NULL)  /* first Q entry */
				fqblock = tbuf;
		}

		tbuf->nblock = NULL;
		tbuf->nuser = NULL;
		tbuf->nqblock = NULL;
		tbuf->fqblock = fqblock;
		tbuf->period = tminutes;
		tbuf->dopt = copts.dopt;
		tbuf->bopt = copts.bopt;
		tbuf->lopt = copts.lopt;
		tbuf->ropt = copts.ropt;
		tbuf->mopt = copts.mopt;
		tbuf->qopt = copts.qopt;
		tbuf->hsnd = 0;
		tbuf->htmout = copts.htmout;
		tbuf->atmout = copts.atmout;
		tbuf->stblock = tbuf == sblock; 
		tbuf->cmain = cmain;
		tbuf->envr = snvr;
		tbuf->ashell = ashell;
		tbuf->ahome = ahome;
		tbuf->anice = anice;
		tbuf->qdelay = 0;
		tbuf->qtmout = qtmout;
		tbuf->tmpbuf = NULL;
		tbuf->exite = 0;
		tbuf->exitc = 0;
		tbuf->swpdelay = 0;
		tbuf->pid = 0;
		tbuf->nextexec = runtab(tbuf, ltm);
		tbuf->lastexec = 0;

		insert_delayjob(tbuf);
		pbuf = tbuf;
	}
	fclose(crontab);

	if(sblock == NULL)
		rmenv(snvr);

	return(sblock);
}

CONFIG *readconfig(char *ctab, char cmain)  /* search for includes and read crontabs */
{
	char *user;
	FILE *crontab;
	char inbuf[1025],*ninbuf,*nptr;
	int i,nentries=0;
	char *ctokens[20];
	CTOPTS copts;
	struct stat buf;
	struct passwd *pword=NULL;
	CONFIG *sblock,*pblock,*lblock,*nblock;

	user = (cmain)?("root"):(ctab);
	copts.bopt = copts.dopt = copts.lopt = copts.ropt = copts.mopt = copts.qopt = 0;
	copts.htmout = copts.atmout = 0;
	copts.mailto = copts.subject = NULL;
	copts.user = user;
	
	sblock = readctab(cpstr(1024, "./%s", ctab), &copts, cmain, &nentries);

	if(! cmain && ! allowinc)
		goto SRET;
	
	if((crontab = fopen(cpstr(1024, "./%s", ctab), "r")) == NULL)
		goto SRET;

	while((pword = getpwnam(user)) == NULL)    /* no such user */
	{
		if(errno != ENOMEM)
		{
			fclose(crontab);
			goto SRET;
		}	
		sleep(1);
	}

	while(fgets(inbuf, 1024, crontab) != NULL)
	{
		ninbuf = inbuf + strspn(inbuf, "\t ");  /* strip leading space */

		if(*ninbuf == '#' || *ninbuf == '\n')
			continue;

		if(strncmp(ninbuf, "INCLUDE ", 8) && strncmp(ninbuf, "include ", 8))
			continue;
		
		ninbuf += 8;
		
		if((nptr = strchr(ninbuf, '\n')) != NULL)
			*nptr = 0;

		if(! *ninbuf || (ctokens[0] = strtok(ninbuf, "\t ")) == NULL)
			continue;
		
		i = 0;
		while(i < 18)
			if(*ctokens[i++] != '-' || (ctokens[i] = strtok(NULL, "\t ")) == NULL)
				break;

		copts.bopt = copts.dopt = copts.lopt = copts.ropt = copts.mopt = copts.qopt = 0;
		copts.htmout = copts.atmout = 0;
		copts.mailto = copts.subject = NULL;
		copts.user = user;

		if(i > 1)
		{
			ctokens[i] = "\0";
			readopts(&ctokens[0], &copts, cmain);
		}	

		ninbuf = ctokens[i-1];

		if(*ninbuf !=  '/' || stat(ninbuf, &buf) < 0 || buf.st_uid != pword->pw_uid ||
		! (buf.st_mode & S_IFREG) || (buf.st_mode & S_IWOTH))
			continue;

		if((pblock = readctab(ninbuf, &copts, cmain, &nentries)) == NULL)
			continue;
				
		if(sblock == NULL)
		{
			sblock = pblock;
		}
		else
		{
			for(lblock = pblock; lblock->nblock != NULL; lblock = lblock->nblock);
			
			sblock->stblock = 0;
			nblock = sblock;
			sblock = pblock;
			lblock->nblock = nblock;
		}
	}	
	fclose(crontab);
	
SRET:
	rmqjobs();
	return(sblock);
}	

void update_ctab(char *ctab)
{
	char cmain=0,*str;
	CONFIG *tbuf;
	int jobs=0;
	struct stat buf;

	if(strcmp(ctab, NCTROOT) == 0)
		cmain = 1;
	
	if(stat(ctab, &buf))  /* crontab removed  */
	{
		njobs -= rmuser(ctab);
		rmqjobs();
		log_info(cpstr(100, "removed %s%s", (cmain)?(ctab):("user "), (cmain)?(""):(ctab)));

		write(svsock, "Crontab removed\n", 16);
		return;
	}

	if((cmain = userperm(ctab)) == -1)
		return;
	
	njobs -= rmuser(ctab);

	if((tbuf = readconfig(ctab, cmain)) != NULL)
	{
		jobs = adduser(tbuf);
		njobs += jobs;
	}

	str = cpstr(100, "Registered %d jobs\n", jobs);
	write(svsock, str, strlen(str));
	
	log_info(cpstr(100, "registered %d jobs for %s%s (total %d)", jobs, (cmain)?("root"):("user "), (cmain)?(""):(ctab), njobs));
}	

void read_config(void)   /* read all crontabs */
{
	DIR *crontabs;
	CONFIG *tbuf,*pbuf;
	int nuser=0;
	char cmain;
	struct dirent *dentry;

	if(chdir(CRONTABS) < 0)
	{
		log_warn("cannot chdir to crontabs: %m");
		return;
	}

	chown(CRONTABS, 0, 0);
	chmod(CRONTABS, 0755);

	if(svctrl != -1)
		close(svctrl);

	if((svctrl = create_socket(SERVCTRL)) == -1)
		log_warn("cannot create socket: %m");

	if((crontabs = opendir(CRONTABS)) == NULL)
	{
		log_warn("cannot open crontab directory: %m");
		return;
	}

	/* remove all crontabs */

	flushqueue();

	tbuf = pbuf = cblock;
	while(tbuf != NULL)
	{
		if(tbuf->stblock)
			rmcstr(tbuf);
			
		pbuf = tbuf;
		tbuf = tbuf->nblock;
		rmtbuf(pbuf);
	}
	cblock = NULL;
	njobs = 0;
	rmqjobs();

	while((dentry = readdir(crontabs)) != NULL)
	{
		if(*dentry->d_name == '.' || *dentry->d_name == '@')
			continue;

		if((cmain = userperm(dentry->d_name)) == -1)	
			continue;

		if((tbuf = readconfig(dentry->d_name, cmain)) != NULL)
		{
			njobs += adduser(tbuf);
			nuser++;
		}
	}
	closedir(crontabs);

	log_info(cpstr(100, "registered %d jobs, %d users", njobs, nuser));
}
